{***********************************
* ViewPort Fuzzy Logic Engine v2.0 *
* (C) 2008 myDanceBot.com, Inc.    *
************************************}{{
Use this Object to add a fuzzy logic engine to your program.
Fuzzy Logic simplifies some control problems by making control values
understandable to humans.
See: http://en.wikipedia.org/wiki/Fuzzy_system

Instead of using complicated formulas coupled with IF statements
for every threshold/exception, fuzzy logic just requires you to
specify how variables should be mapped onto classes.
For example, when trying to park a robot next to a wall,
10cm is too close, 20cm is perfect, and 30cm is too far.
Once fuzzified into classes, the engine supports the standard
functions to let you quickly calculate what you need-
taking less than a millisecond on the Propeller.

Features:
-Easily implement fuzzy logic in your Parallax Propeller programs
-Supports all fuzzy logic operators: not, or, and, fuzzify, defuzzify, andrule, orrule, topclass
-Allows multiple fuzzifying maps and rulesets
-Rich user interface via Viewport: variable membership is shown on maps and rulesets.  All parameters can be changed on-the-fly by clicking the number
-High performance- typical fuzzy logic calculations run in less than 1ms.

Use:
-Include this object in your program:
        Obj f:"fuzzy"
-Set up the mapping functions:
        f.setMap(MapNumber,@VariableToBeMapped,Class4,Class5)
        *The classes will have midpoints at(-Class5,-Class4,0,Class4,Class5)
-Initialize the rules:
        f.setRule(RuleNumber,@Var1,@Var2,@RulePtr)
        *RulePtr should point to an array of 28 bytes
-Register with ViewPort        
        vp.register(f.start(@FuzzFrame))
        *Allocate ~250 longs for the FuzzFrame

In your control loop:
-Fuzzify your variables into the 2 Fuzzy Registers:
        f.fuzzifyA(varA,MapNumber)
        f.fuzzifyB(varB,MapNumber)
-Perform your calculations:
        f.doOr
        f.doInv
        f.doandRule(RuleNumber)
-Defuzzify the result into a variable:
        varC:=f.defuzzify(MapNumber)                
-Share data with ViewPort during development:
        f.share(@FuzzFrame)                                  
}}
CON
  _clkmode        = xtal1 + pll16x
  _xinfreq        = 5_000_000
  muMax=10_000
  maxMap=10
  maxRules=4
var
  long A[5],B[5]          ''array of coefficients of membership into 5 classes
  long A2[3],B2[3]        ''compressed coefficients: 0=#, 1,2 are values
  long Map[4*maxMap]      ''mapping functions
  long RulePtr[maxRules]  ''rulesets
  long MapV[maxMap]       ''rule to variable mapping
  long lvptr
  byte numMaps,numRules
pub start(ap)|n,aptr       ''initializes vp data, returns length of fuzzy data
  aptr:=ap+40
  bytemove(aptr,@numMaps,2)
  aptr+=4
  longmove(aptr,@Map,numMaps*4)
  aptr+=numMaps<<4
  repeat n from 0 to numRules-1
    longmove(aptr,rulePtr[n],7)
    aptr+=28
  longmove(aptr,@MapV,numMaps)
  long[ap]:=44+(numMaps*16)+(numRules*28)+(numMaps*4)
  return ap|3
 
pub share(aptr)|n         ''shares data with viewport at this stage
 
  longmove(aptr,@A,10)   
  aptr+=44
  longmove(@Map,aptr,numMaps<<2)
  aptr+=numMaps<<4      
  repeat n from 0 to numRules-1
    longmove(rulePtr[n],aptr,7)
    aptr+=28
pub setMap(m,vptr,x1,x2)  ''sets up a mapping #m to varn, with points x1,x2
  if lvptr==0
    lvptr:=vptr
  numMaps#>=(m+1)         
  MapV[m]:=(vptr-lvptr)>>2
  m<<=2                   
  Map[m]:=x1
  Map[m+1]:=x2
  Map[m+2]:=x2-x1
pub setVar(vptr)          ''sets reference vptr
  lvptr:=vptr  
pub setRule(r,v1,v2,rptr) ''sets up a rule #r for vars 1,2 with byte array ruleptr
  numRules#>=(r+1)       
  RulePtr[r]:=rptr
  byte[rptr+25]:=(v1-lvptr)>>2
  byte[rptr+26]:=(v2-lvptr)>>2
pub fuzzifyA(x,m)          ''fuzzifies variable m into 5 classes of register A
  fuzzify7(x,False,True,m)
pub fuzzifyB(x,m)          ''fuzzifies variable m into 5 classes of register B
  fuzzify7(x,True,True,m)
pub doAnd|n                ''A=A and B
  repeat n from 0 to 4
    A[n] <#= B[n]
pub doAndRule(r)          ''A=Rule(A and B) using rule m
  doRule(True,r)    
pub doOrRule(r)            ''A=Rule(A or B) using ruleset m
  doRule(False,r)
pub doOr|n                 ''A=A or  B
  repeat n from 0 to 4
    A[n] #>= B[n]
pub doInv|n                ''A=not A
  repeat n from 0 to 4
    A[n]:=muMax-A[n]
pub defuzzify(m):rs        ''defuzzifies A into a number using map m
  m<<=2                   
  rs:=((A[4]-A[0])*Map[m+1]+(A[3]-A[1])*Map[m])/total
pub getTopClass|n,c       ''gets top class for A
  c~
  repeat n from 1 to 4
    if A[n] > A[c]
      c:=n
  return c  
pub get(aptr,doB)         ''gets coefficients, returns ptr
  if doB==1
    longmove(aptr,@B,5)
  else
    longmove(aptr,@A,5)
  return aptr
pub set(aptr,doB)         ''sets coefficients
  if doB==1
    longmove(@B,aptr,5)
  else
    longmove(@A,aptr,5)
pub to5(doB)               ''converts to 5Mode
  if doB
    longfill(@B,0,5)
    B[B2[0]]:=B2[1]
    B[B2[0]+1]:=B2[2]
  else
    longfill(@A,0,5)
    A[A2[0]]:=A2[1]
    A[A2[0]+1]:=A2[2]
pub fuzzifyA2(x,m)         ''fuzzifies variable m into 5 classes of register A-fast mode
  fuzzify7(x,False,False,m)
pub fuzzifyB2(x,m)         ''fuzzifies variable m into 5 classes of register B-fast mode
  fuzzify7(x,True,False,m)     
pub doAndRule2(r)          ''A=Rule(A and B) using rule m- 2Mode
  doRule2(True,r)
pub defuzzify2(m):rs        ''defuzzifies A into a number using map m- 2Mode
  m<<=2                   
  case A2[0]
 
 
    0:rs:=(-A2[1]*Map[m+1]-A2[2]*Map[m])/total2
    1:rs:=-A2[1]*Map[m]/total2
    2:rs:=A2[2]*Map[m]/total2
    3:rs:=(A2[2]*Map[m+1]+A2[1]*Map[m])/total2
pri doRule(andM,r)|t[5],x,y,v,tidx
  tIdx:=ruleptr[r]
  longmove(@t,@A,5)
  longfill(@A,0,5)
  if andM
    repeat y from 0 to 4
      if B[y]>0
        repeat x from 0 to 4
          v:=t[x] <# B[y]
          if v > 0
            A[byte[tIdx]] #>= v   
          tidx+=1
      else                      
        tidx+=5            
  else
    repeat y from 0 to 4
      if B[y]>0
        repeat x from 0 to 4
          v:=t[x] #> B[y]
          if v > 0
            A[byte[tIdx]] #>= v   
          tidx+=1
      else            
        tidx+=5              
pri doRule2(andM,r)|tIdx,v0,v1,v2,v3,tv,zT
  longfill(@A,0,5)
  tIdx:=ruleptr[r]+(B2[0])*5+A2[0]  
  if andM
    v0:=A2[1] <# B2[1]
    v1:=A2[2] <# B2[1]
    v2:=A2[1] <# B2[2]
    v3:=A2[2] <# B2[2]
  else
    v0:=A2[1] #> B2[1]
    v1:=A2[2] #> B2[1]
    v2:=A2[1] #> B2[2]
    v3:=A2[2] #> B2[2]
  if v0 > 0
    tv:=byte[tidx]
    A[tv]:=v0
  tIdx+=1    
  if v1 > 0
    tV:=byte[tidx]  
    A[tV] #>= v1
  tIdx+=4     
  if v2 > 0
    tV:=byte[tIdx] 
    A[tV] #>= v2
  tIdx+=1    
  if v3 > 0
    tV:=byte[tIdx] 
    A[tV] #>= v3
pri total:t
  t:=A[0]+A[1]+A[2]+A[3]+A[4]
pri total2
  return A2[1]+A2[2]
pub fuzzify7(x,doB,do5,m)|x0,x1,x2
  {{fuzzifies x into A or B with 5 or 3 members}}
 
  m<<=2                   
  if x > 0
    if x > Map[m+1]
      x0:=3 
      x1~
      x2:=muMax
    elseif x > Map[m]
      x0:=3
      x1:=(Map[m+1]-x) * muMax/Map[m+2]
      x2:=muMax-x1
    else
      x0:=2
      x2:=x* muMax/Map[m]
      x1:=muMax-x2
  else      
    if x < -Map[m+1]
      x0~ 
      x1:=muMax
      x2~
    elseif x < -Map[m]
      x0~
      x2:=(x+Map[m+1]) * muMax/Map[m+2]
      x1:=muMax-x2
    else
      x0:=1
      x1:=-x* muMax/Map[m]
      x2:=muMax-x1
  if do5
    if doB
      longfill(@B,0,5)
      B[x0]:=x1
      B[x0+1]:=x2
    else
      longfill(@A,0,5)
      A[x0]:=x1
      A[x0+1]:=x2
  else
    if doB
      B2[0]:=x0
      B2[1]:=x1
      B2[2]:=x2
    else
      A2[0]:=x0
      A2[1]:=x1
      A2[2]:=x2
                                                                                                     